
/**
 * This is the third application used in the Java-GNOME tutorial.
 * It demonstrates the use of toolbars and statusbars.
 */

import gtk.*;
import gnome.*;

public class Third {
    private GnomeApp app = null;
    private GtkToolbar toolbar = null;
    private GtkStatusbar statusbar = null;
    private int contextID = 0;
    public static final String appVersion = "0.1";

    public Third() {
	createMainWindow();
	createMenus();
	createToolbar();
	createStatusbar();
    }

    private void createMainWindow() {
	app = new GnomeApp("Third", "Third App");
	app.setPolicy(false, true, false);
	app.setWmclass("Third", "ThirdApp");
	app.setUsize(200, 200);
	app.signal_connect("delete_event", this);
	app.signal_connect("destroy", this);
	app.show();
    }

    private void createMenus() {
	GtkMenuBar menubar = new GtkMenuBar();
	GtkMenuItem file = new GtkMenuItem("File");
	GtkMenuItem help = new GtkMenuItem("Help");
	menubar.append(file);
	menubar.append(help);
	
	// finish the file menu
	GtkMenu fileMenu = new GtkMenu();
	GtkMenuItem fileNew = new GtkMenuItem("New");
	GtkMenuItem fileOpen = new GtkMenuItem("Open");
	GtkMenuItem fileSave = new GtkMenuItem("Save");
	GtkMenuItem fileSaveAs = new GtkMenuItem("Save As");
	GtkMenuItem fileExit = new GtkMenuItem("Exit");
	fileNew.signal_connect("activate", "fileNew", this);
	fileOpen.signal_connect("activate", "fileOpen", this);
	fileSave.signal_connect("activate", "fileSave", this);
	fileSaveAs.signal_connect("activate", "fileSaveAs", this);
	fileExit.signal_connect("activate", "fileExit", this);
	file.setSubmenu(fileMenu);
	fileMenu.append(fileNew);
	fileMenu.append(fileOpen);
	fileMenu.append(fileSave);
	fileMenu.append(fileSaveAs);
	fileMenu.append(fileExit);
	// now display it
	fileNew.show();
	fileOpen.show();
	fileSave.show();
	fileSaveAs.show();
	fileExit.show();
	file.show();
	
	// finish the help menu
	GtkMenu helpMenu = new GtkMenu();
	GtkMenuItem helpAbout = new GtkMenuItem("About");
	helpAbout.signal_connect("activate", "helpAbout", this);
	help.setSubmenu(helpMenu);
	helpMenu.append(helpAbout);
	// display
	helpAbout.show();
	help.show();
	
	menubar.show();
	
	app.setMenus(menubar);
    }

    private void createToolbar() {
	toolbar = new GtkToolbar(GtkOrientation.GTK_ORIENTATION_VERTICAL,
				 GtkToolbarStyle.GTK_TOOLBAR_TEXT);
	GtkButton button = new GtkButton("New");
	button.signal_connect("clicked", "fileNew", this);
	toolbar.appendWidget(button, "New File", "Private");
	button.show();
	button = new GtkButton("Open");
	button.signal_connect("clicked", "fileOpen", this);
	toolbar.appendWidget(button, "Open File", "Private");
	button.show();
	toolbar.appendSpace();
	button = new GtkButton("Save");
	button.signal_connect("clicked", "fileSave", this);
	toolbar.appendWidget(button, "Save File", "Private");
	button.show();
	button = new GtkButton("Save As");
	button.signal_connect("clicked", "fileSaveAs", this);
	toolbar.appendWidget(button, "Save File As", "Private");
	button.show();

	app.setToolbar(toolbar);
    }

    private void createStatusbar() {
	statusbar = new GtkStatusbar();
	contextID = statusbar.getContextId("Third App");
	statusbar.show();
	statusbar.push(contextID, "Ready");

	app.setStatusbar(statusbar);
    }

    public boolean delete_event(int value) {
	return true;
    }
    
    public void destroy() {
	Gtk.mainQuit();
    }
    
    public void fileNew() {
	statusbar.pop(contextID);
	statusbar.push(contextID, "File New Selected");
    }

    public void fileOpen() {
	statusbar.pop(contextID);
	statusbar.push(contextID, "file Open Selected");
    }

    public void fileSave() {
	statusbar.pop(contextID);
	statusbar.push(contextID, "File Save Selected");
    }

    public void fileSaveAs() {
	statusbar.pop(contextID);
	statusbar.push(contextID, "File Save As Selected");
    }

    public void fileExit() {
	Gtk.mainQuit();
    }
    
    public void helpAbout() {
	String title = "Third App";
	String version = "Version " + appVersion;
	String license = "GPL";
	String[] authors = {"The Java-GNOME team"};
	String comments = "This example is a part of the Java-GNOME tutorial";
	String pixmap = "gftp.png";
	GnomeAbout about = new GnomeAbout(title, version, license,
					  authors, comments, pixmap);
	about.show();
    }
    
    public static void main(String[] args) {
	Gnome.init("Third", Third.appVersion, 0, 0);
	Third second = new Third();
	Gtk.main();
    }    
}
