/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/


import java.io.*;
import java.util.*;

class ClassDefinition {
    TypeDefinition classType;
    String packageName;
    String className;
    String inherit;
    Vector enums = new Vector();
    Vector fields = new Vector();
    Vector methods = new Vector();
    Vector constructors = new Vector();
    DefaultConstructorDefinition defaultcsd = null;
    Vector imports;
    
    public void addConstructor(ConstructorDefinition c) {
	constructors.addElement(c);
    }

    public void addEnum(EnumDefinition e) {
	enums.addElement(e);
    }

    public void addField(FieldDefinition e) {
	fields.addElement(e);
    }

    public void addMethod(MethodDefinition m) {
	methods.addElement(m);
    }

    void emitFileHeader(PrintWriter out, String filename) {
	out.println("/*");
	out.println(" *   Java-Gnome");
	out.println(" *");
	out.println(" *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>");
	out.println(" *");
	out.println(" *   The Java-Gnome bindings library is free software distributed under");
	out.println(" *   the terms of the GNU Library General Public License version 2 ");
	out.println(" *");
	out.println(" ***********************************************************************/");
	out.println();
	out.println("/* Java-Gnome bindings -- " + filename + " -- " + new Date() + " */");
	out.println("/* DO NOT EDIT - This file has been automagically generated */");
	out.println();
    }

    public String getClassName() {
	return className;
    }

    public TypeDefinition getClassType() {
	return classType;
    }

    public ClassDefinition(String p, String c, String e) {
	packageName = p;
	packageName.toLowerCase();
	className = c;
	inherit = e;
    }

    public ClassDefinition(String p, Vector i, String c, String e) {
	this(p, c, e);
	imports = (Vector) i.clone();
	imports.removeElement(packageName);
    }

    public ClassDefinition(String p, Vector i, TypeDefinition c, TypeDefinition e) {
	packageName = p;
	packageName.toLowerCase();
	imports = (Vector) i.clone();
	imports.removeElement(packageName);
	classType = c;
	className = c.getJavaType();
	if (e != null)
	    inherit = e.getJavaType();
    }

    public void saveCGlue() {
	try {
	    String filename = packageName.toLowerCase() + "_" + className;
	    // FIXME: no "glue" directory name hardcoded
	    String filepath = "glue" + File.separator + filename + ".c";
	    PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filepath)));
	    emitFileHeader(out, filepath);
	    out.println("#include \"" + filename + ".h\"");
	    if (!packageName.toLowerCase().equals("gdk")) {
		out.println("#include \"../other/" + packageName.toLowerCase() + "_wrappers.h\"");
	    }
	    out.println("#ifdef __cplusplus");
	    out.println("extern \"C\" {");
	    out.println("#endif");
	    for (Enumeration e = constructors.elements(); e.hasMoreElements();) {
		((ConstructorDefinition) e.nextElement()).printNativeFunction(out);
	    }
	    if (defaultcsd != null)
		defaultcsd.printNativeFunction(out);
	    for (Enumeration e = fields.elements(); e.hasMoreElements();) {
		((FieldDefinition) e.nextElement()).printNativeFunction(out);
	    }
	    for (Enumeration e = methods.elements(); e.hasMoreElements();) {
		((MethodDefinition) e.nextElement()).printNativeFunction(out);
	    }
	    out.println("#ifdef __cplusplus");
	    out.println("}");
	    out.println("#endif");
	    out.close();
	}
	catch (IOException e) {
	    System.err.println(e);
	}
    }

    public void saveCHeader() {
	try {
	    String fileName = packageName.toLowerCase() + "_" + className;
	    // FIXME: no "glue" directory name hardcoded
	    String filepath = "glue" + File.separator + fileName + ".h";
	    PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filepath)));
	    emitFileHeader(out, filepath);
	    out.println("/* DO NOT EDIT THIS FILE - it is machine generated */");
	    out.println("/* by Java/Gnome Fake JNI Generator */");
	    out.println("#include <jni.h>");
	    out.println("#include <sys/types.h>");
	    if (Parser.currentModule.equals("Gtk"))
		out.println("#include <gtk/gtk.h>");
	    if (Parser.currentModule.equals("Gdk"))
		out.println("#include <gdk/gdk.h>");
	    if (Parser.currentModule.equals("Gnome"))
		out.println("#include <gnome.h>");
	    out.println("/* Header for class " + fileName + " */");
	    out.println();
	    out.println("#ifndef _Included_" + fileName);
	    out.println("#define _Included_" + fileName);
	    out.println("#ifdef __cplusplus");
	    out.println("extern \"C\" {");
	    out.println("#endif");
	    for (Enumeration e = constructors.elements(); e.hasMoreElements();) {
		((ConstructorDefinition) e.nextElement()).emitNativeHeader(out);
	    }
	    if (defaultcsd != null)
		defaultcsd.emitNativeHeader(out);
	    for (Enumeration e = fields.elements(); e.hasMoreElements();) {
		((FieldDefinition) e.nextElement()).emitNativeHeader(out);
	    }
	    for (Enumeration e = methods.elements(); e.hasMoreElements();) {
		((MethodDefinition) e.nextElement()).emitNativeHeader(out);
	    }
	    out.println("#ifdef __cplusplus");
	    out.println("#endif");
	    out.println("#endif");
	    out.close();
	}
	catch (IOException e) {
	    System.err.println(e);
	}
    }

    public void saveFile() {
	try {
	    String filepath = packageName.toLowerCase() + File.separator + className + ".java";
	    PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filepath)));
	    emitFileHeader(out, filepath);
	    out.println("package " + packageName.toLowerCase() + ";");
	    out.println();
	    if (imports != null)
		for (Enumeration e = imports.elements(); e.hasMoreElements();) {
		    String pck = (String) e.nextElement();
		    out.print("import " + pck.toLowerCase() + ".*;\n");
		}
	    out.println("import java.util.*;");
	    out.print("\n");
	    if (inherit != null)
		out.print("public class " + className + " extends " + inherit);
	    else
		out.print("public class " + className);
	    out.print("{\n\n");
	    
	    /* This section adds the System.loadLibrary to the Gtk, Gnome and Gdk
	       classes so that each programs does not have to do it for us.
	       It is necessary for Gtk.java and Gnome.java (both tested) but also
	       for Gdk.java (I think). */
	    if (className.equals("Gdk") || className.equals("Gtk")) {
		out.print("  static\n    {\n");
		out.print("      System.loadLibrary(\"GTKJava\");\n");
		out.print("    }\n\n");
	    }
	    else {
		if (className.equals("Gnome")) {
		    out.print("  static\n    {\n");
		    out.print("      System.loadLibrary(\"GNOMEJava\");\n");
		    out.print("    }\n\n");
		}
		else {
		    out.println("  public " + className + "(long _peer) {");
		    if (inherit != null)
			out.println("     nativepeer = _peer;");
		    out.println("  }");
		    out.println("");
		}
	    }
	    ConstructorDefinition cst = null;
	    boolean hasdefault = false;
	    for (Enumeration e = constructors.elements(); e.hasMoreElements();) {
		cst = (ConstructorDefinition) e.nextElement();
		if (cst.isDefault())
		    hasdefault = true;
		cst.emitJavaMethod(out);
		if (cst.hasDefault())
		    hasdefault = true;
		out.println();
	    }
	    // We *must* have a no-arg constructor
	    if ((!hasdefault) && (constructors.size() > 0)) {
		String prefix = cst.nativeName.substring(0, cst.nativeName.indexOf("_new"));
		defaultcsd = new DefaultConstructorDefinition(prefix);
		defaultcsd.objectType = cst.objectType;
		defaultcsd.objectName = cst.objectName;
		defaultcsd.packageName = cst.packageName;
		defaultcsd.methodName = "nativeinit";
		//System.err.print(">->->"+prefix);
		defaultcsd.emitJavaMethod(out);
		out.println();
		/*out.println("  public "+className+"()");
		  out.println("  {");
		  out.println("      ");
		  out.println("  }");*/
	    }
	    if ((!hasdefault) && (constructors.size() == 0)) {
		out.println("  public " + className + "() {");
		out.println("      ");
		out.println("  }");
	    }
	    for (Enumeration e = enums.elements(); e.hasMoreElements();) {
		EnumDefinition enum = (EnumDefinition) e.nextElement();
		enum.emit(out);
		out.print("\n");
	    }
	    out.print("\n");
	    for (Enumeration e = fields.elements(); e.hasMoreElements();) {
		((FieldDefinition) e.nextElement()).emitJavaField(out);
	    }
	    out.print("\n");
	    for (Enumeration e = methods.elements(); e.hasMoreElements();) {
		((MethodDefinition) e.nextElement()).emitJavaMethod(out);
	    }
	    out.print("\n}\n");
	    out.close();
	    saveCHeader();
	    saveCGlue();
	}
	catch (IOException e) {
	    System.err.println(e);
	}
    }
}
