/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/


import java.util.*;
import java.io.*;

public class ConstructorDefinition extends MethodDefinition {
    String nativePreGlue = null;
    String nativeGlue = null;
    String nativePostGlue = null;
    
    public void emitJavaDefaultStubs(PrintWriter pw, int arguments, int defaultValues) {
	// This section create the methods for the methods who has default values
	if (defaultValues > 0)
	    pw.println("\n  /* " + objectName + " has " + defaultValues + 
		       " default argument value(s) */");
	for (int i = 1; i <= defaultValues; i++) {
	    // First writes the prototype(stub) method
	    int defaultValueNo = 0;
	    pw.print("  ");
	    if (isStatic)
		pw.print("static ");
	    pw.print("public ");
	    pw.print(objectName + "(");
	    Enumeration e = args.elements();
	    boolean firstIsDefault = false;
	    Argument a = (Argument) e.nextElement();
	    if (!(a.hasDefaultValue() && i == defaultValues))
		pw.print(a.toString());
	    else
		firstIsDefault = true;
	    if (a.hasDefaultValue())
		defaultValueNo++;
	    for (int j = 1; j < arguments; j++) {
		a = (Argument) e.nextElement();
		if (a.hasDefaultValue())
		    defaultValueNo++;
		if (!(a.hasDefaultValue() && defaultValueNo > defaultValues - i)) {
		    if (firstIsDefault)
			firstIsDefault = false;
		    else
			pw.print(", ");
		    pw.print(a.toString());
		}
	    }
	    pw.println(") {");
	    
	    // Then write the call to the real method with the default arguments
	    defaultValueNo = 0;
	    pw.print("    ");
	    pw.print("nativepeer = " + methodName + "(");
	    e = args.elements();
	    a = (Argument) e.nextElement();
	    if (!(a.hasDefaultValue() && i == defaultValues))
		pw.print(a.getName());
	    else
		pw.print(a.getDefaultValue());
	    if (a.hasDefaultValue())
		defaultValueNo++;
	    for (int j = 1; j < arguments; j++) {
		a = (Argument) e.nextElement();
		if (a.hasDefaultValue())
		    defaultValueNo++;
		pw.print(", ");
		if (!(a.hasDefaultValue() && defaultValueNo > defaultValues - i))
		    pw.print(a.getName());
		else
		    pw.print(a.getDefaultValue());
	    }
	    pw.println(");\n  }\n");
	}
	pw.println("");
    }

    public void emitJavaMethod(PrintWriter pw) {
	/*pw.println("  "+objectName+"(long _peer)");
	  pw.println("  {");
	  pw.println("     nativepeer = _peer;");
	  pw.println("  }");
	  pw.println("");
	*/
	pw.print("  public " + objectName + "(");
	int defaultValues = 0;
	int arguments = 0;
	Enumeration e = args.elements();
	if (e.hasMoreElements()) {
	    Argument a = (Argument) e.nextElement();
	    if (a.hasDefaultValue())
		defaultValues++;
	    pw.print(a.toString());
	    for (; e.hasMoreElements();) {
		a = (Argument) e.nextElement();
		if (a.hasDefaultValue())
		    defaultValues++;
		arguments++;
		pw.print(", ");
		pw.print(a.toString());
	    }
	}
	pw.println(") {");
	pw.print("     nativepeer = " + methodName + "(");
	e = args.elements();
	if (e.hasMoreElements()) {
	    pw.print(((Argument) e.nextElement()).getName());
	    for (; e.hasMoreElements();) {
		pw.print(", ");
		pw.print(((Argument) e.nextElement()).getName());
	    }
	}
	pw.println(");");
	pw.println("  }");
	
	// let the superclass emit the native declaration and wrapper
	// method (if needed), as well as call back into the calss
	// to emit the default-arg variants
	super.emitJavaMethod(pw);
    }

    public void emitNativeBody(PrintWriter pw) {
	//      System.err.println("NATIVE CONSTRUCTOR");
	
	int i = 0;
	if (nativePreGlue != null) {
	    pw.print(nativePreGlue + '\n');
	}
	
	/* Emit local variables declarations */
	for (Enumeration e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitNativeDeclaration(pw, "_arg" + i);
	    i++;
	}
	pw.print("  " + objectType.getNativeType());
	pw.println(" _result;");
	pw.print("  " + returnType.getJNIType());
	pw.println(" _jresult;");
	pw.println();
	
	/* Argument translation */
	i = 0;
	for (Enumeration e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitJavaToNativeConversion(pw, "_arg" + i);
	    i++;
	}
	i = 0;
	Enumeration e = args.elements();
	if (nativeGlue != null) {
	    pw.print(nativeGlue + '\n');
	}
	else {
	    pw.print("  ");
	    pw.print("_result = (" + objectType.getNativeType() + ")");
	    pw.print(nativeName + "(");
	    if (e.hasMoreElements()) {
		pw.print("_arg" + i);
		e.nextElement();
		i++;
	    }
	    for (; e.hasMoreElements();) {
		e.nextElement();
		pw.print(", _arg" + i);
		i++;
	    }
	    pw.println(");");
	}
	i = 0;
	for (e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitNativeToJavaConversion(pw, "_arg" + i);
	    i++;
	}
	if (nativePostGlue != null) {
	    pw.print(nativePostGlue + '\n');
	}
	pw.println();
	pw.println("  *(" + objectType.getNativeType() + "*)&_jresult = _result;");
	pw.println("  return _jresult;");
    }

    /**
     * Get the value of nativeGlue.
     * @return Value of nativeGlue.
     */
    public String getNativeGlue() {
	return nativeGlue;
    }

    public boolean hasDefault() {
	int defaultValues = 0;
	Enumeration e = args.elements();
	if (e.hasMoreElements()) {
	    if (((Argument) e.nextElement()).hasDefaultValue())
		defaultValues++;
	    for (; e.hasMoreElements();) {
		if (((Argument) e.nextElement()).hasDefaultValue())
		    defaultValues++;
	    }
	}
	return args.size() - defaultValues == 0;
    }

    public boolean isDefault() {
	return args.size() == 0;
    }

    public ConstructorDefinition(String n, Vector a) {
	super(n, new TypeDefinition(TypeDefinition.LONG), a);
    }

    public void setNativeGlue(String v) {
	this.nativeGlue = v;
    }

    public void setNativePostGlue(String v) {
	this.nativePostGlue = v;
    }

    /**
     * Set the value of nativeGlue.
     * @param v  Value to assign to nativeGlue.
     */
    public void setNativePreGlue(String v) {
	this.nativePreGlue = v;
    }

    public boolean translate(String prefix, TypeDefinition obj) {
	objectName = obj.getRawValue();
	String m = convertFunc(nativeName, Parser.currentModule, obj.getJavaType());
	if (m != null) {
	    methodName = "native" + m;
	    objectType = obj;
	    objectName = obj.getJavaType();
	    packageName = prefix.toLowerCase();
	    return true;
	}
	else
	    return false;
    }
}
