/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/

import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.io.*;

public class EnumDefinition {
    String packageName;
    String enumName;
    Vector identifiers;

    void emitFileHeader(PrintWriter out, String filename) {
	out.println("/*");
	out.println(" *   Java-Gnome");
	out.println(" *");
	out.println(" *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>");
	out.println(" *");
	out.println(" *   The Java-Gnome bindings library is free software distributed under");
	out.println(" *   the terms of the GNU Library General Public License version 2 ");
	out.println(" *");
	out.println(" ***********************************************************************/");
	out.println();
	out.println("/* Java-Gnome bindings -- " + filename + " -- " + new Date() + " */");
	out.println("/* DO NOT EDIT - This file has been automagically generated */");
	out.println();
    }
    
    public void emit(PrintWriter pw) {
	String name = getName();
	pw.println("  static private " + name + "[] theInterned = {");

	StringBuffer sb = new StringBuffer();
	int base = 0;
	Enumeration e = identifiers.elements();
	while (e.hasMoreElements()) {
	    // spit out the array elements directly
	    if (base != 0)
		pw.println(",");
	    pw.print("   new ");
	    pw.print(name);
	    pw.print("(");
	    pw.print(base);
	    pw.print(")");

	    // accumulate other static definitions for later
	    String s = ((TypeDefinition) e.nextElement()).translateNaming();
	    sb.append("  public final static int _");
	    sb.append(s);
	    sb.append(" = ");
	    sb.append(base);
	    sb.append(";\n");
	    sb.append("  public final static ");
	    sb.append(name);
	    sb.append(" ");
	    sb.append(s);
	    sb.append(" = theInterned[");
	    sb.append(base);
	    sb.append("];\n");

	    base++;
	}

	pw.println("\n  };");

	// the aforementioned "later"
	pw.print(sb.toString());
    }
    
    public String getName() {
	return enumName;
    }
    
    public EnumDefinition(String p, String n, Vector a) {
	packageName = p;
	packageName.toLowerCase();
	enumName = n;
	identifiers = a;
    }
    
    public EnumDefinition(String n, Vector a) {
	enumName = n;
	identifiers = a;
    }
    
    public void saveFile() {
	try {
	    System.err.println("savefile " + packageName + "/" + getName());
	    String name = getName();
	    String filepath = packageName.toLowerCase() + File.separator + getName() + 
		".java";
	    PrintWriter out = new PrintWriter(new BufferedWriter(
								 new FileWriter(filepath)));
	    emitFileHeader(out, filepath);

	    out.println("package " + packageName.toLowerCase() + ";\n" + 
			"\n" +
			"import java.util.Hashtable;\n" +
			"\n" +
			"final public class " + name + " {\n" +
			"  int value_;\n" +
			"\n" +
			"  public " + name + "(int value) {\n" +
			"    value_ = value;\n" +
			"  }\n");

	    out.println("  public int getValue() {\n" +
			"    return value_;\n" +
			"  }\n");

	    out.println("  public int hashCode() {\n" +
			"    return value_;\n" +
			"  }\n");

	    out.println("  public boolean equals(Object other) {\n" +
			"    return (other instanceof " + name + ")\n" +
			"      && (((" + name + ") other).value_ == value_);\n" +
			"  }\n");

	    out.println("  static private Hashtable theInternedExtras = null;\n" +
			"  static private " + name + " theSacrificialOne = new " +
			name + "(0);\n" +
			"  static public synchronized " + name +
			" intern(int value) {\n" +
			"    if (value < theInterned.length)\n" +
			"      return theInterned[value];\n" +
			"    theSacrificialOne.value_ = value;\n" +
			"    if (theInternedExtras == null)\n" +
			"      theInternedExtras = new Hashtable();\n" +
			"    " + name + " already  = (" + name +
			") theInternedExtras.get(theSacrificialOne);\n" +
			"    if (already == null) {\n" +
			"      already = new " + name + "(value);\n" +
			"      theInternedExtras.put(already, already);\n" +
			"    }\n" +
			"    return already;\n" +
			"  }");


	    emit(out);
	    out.println("}");
	    out.close();
	}
	catch (IOException e) {
	    System.err.println(e);
	}
    }
}
