/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998-1999 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/

import java.io.*;
import java.util.*;

/** The main parser code has not been rewritten yet, so it is still
    full of hacks, and is definitely not optimized */

public class Parser {
    public static final boolean debug = false;


    /** This field controls naming style. If true, the parser will adopt
	Java style naming, and if not, it will stick to classic C
	naming. For instance gtk_window_set_title would be translated
	into gtk.Window.setTitle(...) in PURE_JAVA style,
	gtk.GtkWindow.set_title(...) in PURE_C style and
	gtk.GtkWindow.setTitle(...) in so-called MIXED style. A quite
	wide consensus is on mixed style, so I am adopting this */
    
    public final static int PURE_C = 1;
    public final static int PURE_JAVA = 2;
    public final static int MIXED = 3;
    public static int naming = MIXED;
    
    /* repositories for identifiers found in defs files */
    Vector classnames = new Vector();
    Vector methodnames = new Vector();
    Vector constantnames = new Vector();
    Vector enumsnames = new Vector();
    Vector typesnames = new Vector();
    Vector allenumsnames = new Vector();
    /* modules known (i.e. gdk, gtk, gnome) */
    static Vector modules = new Vector();
    
    // these one are hackish. I should move to a pure ooc parser
    static String currentModule;
    static String currentPackage;
    
    public void addClassDefinition(ClassDefinition cd) {
	/* lexicographic order is quite important in the class vector,
	   as it allows correct function to method name conversion in
	   case of classes sharing common prefixes (i.e. GtkMenu and
	   GtkMenuItem) */
	/* I should also rewrite this with the Java Collection package in mind */
	
	boolean first = false;
	String ref = cd.getClassName();
	for (Enumeration e = classnames.elements(); e.hasMoreElements();) {
	    String c = ((ClassDefinition) e.nextElement()).getClassName();
	    if (ref.startsWith(c))
		first = true;
	}
	if (first)
	    classnames.insertElementAt(cd, 0);
	else
	    classnames.addElement(cd);
    }
    
    public void addConstructorDefinition(ConstructorDefinition cd) {
	constantnames.addElement(cd);
    }

    public void addEnumDefinition(EnumDefinition e) {
	enumsnames.addElement(e);
    }
    
    public void addMethodDefinition(MethodDefinition md) {
	methodnames.addElement(md);
    }
    
    public void addTypeDefinition(TypeDefinition td) {
	if (td.isObject()) {
	    for (Enumeration e = enumsnames.elements(); e.hasMoreElements();) {
		EnumDefinition t = (EnumDefinition) e.nextElement();
		if (t.getName().endsWith(td.getRawValue()))
		    td.setEnumValue();
	    }
	}
	if (!typesnames.contains(td))
	    typesnames.addElement(td);
    }
    
    public void checkTypes(String name) {
	boolean replaced = false;
	if ((name.toLowerCase()).startsWith(currentModule.toLowerCase()))
	    name = name.substring(currentModule.length());
	for (Enumeration e = typesnames.elements(); e.hasMoreElements();) {
	    TypeDefinition t = (TypeDefinition) e.nextElement();
	    if (t.isObject()) {
		if (t.getRawValue().equals(name)) {
				t.setEnumValue();
				replaced = true;
		}
	    }
	}
    }
    
    /*
      public void skipLPar(StreamTokenizer st)
      {
      if (st.ttype == (int) '(')
      st.nextToken();
      }
      public String nextPureToken(StreamTokenizer st)
      {
      String lexeme = st.nextToken();
      
      while (lexeme.lastOf().equals(")"))
      {
      System.err.println("*** Detected ')' - should pushback");
      lexeme = lexeme.substring(0,lexeme.length()-1);
      }
      return lexeme;
      }
    */
    
    public void getAllEnums(String enum_file) {
	try {
	    BufferedReader br = new BufferedReader(new InputStreamReader(
									 new FileInputStream(enum_file)));
	    StreamTokenizer ste = new StreamTokenizer(br);
	    ste.eolIsSignificant(true);
	    while (ste.ttype != ste.TT_EOF) {
		ste.nextToken();
		if ((ste.ttype == ste.TT_WORD))
		    allenumsnames.addElement(ste.sval);
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println(e);
	}
    }
    
    public TypeDefinition getTypeDefinition(String s) {
	TypeDefinition td = new TypeDefinition();
	td = TypeDefinition.convertFromScheme(s);
	if (td.isObject()) {
	    for (Enumeration e = allenumsnames.elements(); e.hasMoreElements();) {
		String enums = (String) e.nextElement();
		if (enums.endsWith(s))
		    td.setEnumValue(s);
	    }
	}
	return td;
    }
    
    public static void main(String[] args) {
	try {
	    currentModule = args[1];
	    currentPackage = currentModule.toLowerCase();
	    modules.addElement(currentModule);
	    Parser p = new Parser();
	    p.parseFile(args[0] + File.separator + args[1].toLowerCase() + ".defs");
	}
	catch (Exception e) {
	    System.err.println(e.getMessage());
	    e.printStackTrace();
	}
    }
    
    public void parseDefineBoxed(Vector v) {
	TypeDefinition boxedName;
	TypeDefinition inheritName;
	try {
	    boxedName = getTypeDefinition((String) (v.elementAt(1)));
	    if (!boxedName.getRawValue().equals("GtkOjbect")) {
		//inheritName = 
		//getTypeDefinition((String)(((Vector)(v.elementAt(2))).firstElement()));
		// addTypeDefinition(inheritName);
		
		addTypeDefinition(boxedName);
		ClassDefinition cdf = new ClassDefinition(currentModule, 
							  modules, boxedName, null);
		//inheritName);
		addClassDefinition(cdf);
	    }
	}
	catch (Exception e) {
	    System.err.println("parse define-boxed error: " + e.toString() + " " + v);
	    e.printStackTrace();
	}
    }
    
    public void parseDefineEnum(Vector v) {
	String enumName;
	try {
	    Enumeration e = v.elements();
	    e.nextElement();
	    enumName = (String) e.nextElement();
	    checkTypes(enumName);
	    Vector a = new Vector();
	    for (; e.hasMoreElements();) {
		Vector args = (Vector) e.nextElement();
		if (debug)
		    System.err.println("    enum-item:" + args);
		Enumeration e2 = args.elements();
		for (; e2.hasMoreElements();) {
		    e2.nextElement(); // removing enum scheme name
		    TypeDefinition td = new TypeDefinition();
		    td.setEnumValue((String) e2.nextElement());
		    addTypeDefinition(td);
		    if (debug)
			System.err.println("         item-rawvalue:" + td.getRawValue());
		    a.addElement(td);
		}
	    }
	    EnumDefinition cdf = new EnumDefinition(currentModule, enumName, a);
	    if (debug)
		System.err.println("  enum:" + enumName + " " + a);
	    addEnumDefinition(cdf);
	}
	catch (Exception e) {
	    System.err.println("parse define-enum error: " + e.toString() + " " + v);
	}
    }
    
    public void parseDefineFlags(Vector v) {
	String flagsName;
	try {
	    Enumeration e = v.elements();
	    e.nextElement();
	    flagsName = (String) e.nextElement();
	    checkTypes(flagsName);
	    Vector a = new Vector();
	    for (; e.hasMoreElements();) {
		Vector args = (Vector) e.nextElement();
		Enumeration e2 = args.elements();
		for (; e2.hasMoreElements();) {
		    //Vector arg = (Vector)e2.nextElement();
		    //Enumeration e3 = arg.elements();
		    
		    e2.nextElement(); // removing enum scheme name
		    TypeDefinition td = new TypeDefinition();
		    td.setEnumValue((String) e2.nextElement());
		    addTypeDefinition(td);
		    a.addElement(td);
		}
	    }
	    FlagsDefinition cdf = new FlagsDefinition(currentModule, flagsName, a);
	    //if (debug)
	    System.err.println("flags:" + flagsName + " " + a);
	    addEnumDefinition(cdf);
	    //  System.err.println(
	    
	}
	catch (Exception e) {
	    System.err.println("parse define-enum error: " + e.toString() + " " + v);
	}
    }
    
    public void parseDefineFunc(Vector v) {
	String metName;
	String native_pre = null;
	String native_call = null;
	String native_post = null;
	TypeDefinition retType;
	try {
	    Enumeration e = v.elements();
	    e.nextElement();
	    metName = ((String) e.nextElement()).toLowerCase();
	    boolean isConstructor = (metName.indexOf("new") != -1);
	    retType = getTypeDefinition((String) e.nextElement());
	    addTypeDefinition(retType);
	    Vector args = (Vector) e.nextElement();
	    Enumeration e2 = args.elements();
	    Vector a = new Vector();
	    for (; e2.hasMoreElements();) {
		Vector arg = (Vector) e2.nextElement();
		Enumeration e3 = arg.elements();
		String s = (String) e3.nextElement();
		TypeDefinition argtype = getTypeDefinition(s); //(String)e3.nextElement());
		if (debug)
		    System.err.println("---argtypedef:" + s + argtype);
		
		// FIXME: insert here default arguments and null-ok processing code
		addTypeDefinition(argtype);
		s = (String) e3.nextElement();
		Argument argument = new Argument(argtype, s);
		if (debug)
		    System.err.println("---argname:" + s);
		if (e3.hasMoreElements()) {
		    Vector defaultValue = (Vector) e3.nextElement();
		    if (((String) (defaultValue.elementAt(0))).equals("="))
			argument.setDefaultValue((String) (defaultValue.elementAt(1)));
		    if (((String) (defaultValue.elementAt(0))).equals("null-ok"))
			argument.setDefaultValue("null");
		}
		a.addElement(argument);
	    }
	    if (e.hasMoreElements()) {
		if (debug)
		    System.err.println("---NATIVEGLUE:" + e);
		Vector jniCode = (Vector) e.nextElement();
		if (((String) (jniCode.elementAt(0))).equals("jni-code")) {
		    native_pre = ((String) (jniCode.elementAt(1)));
		    native_call = ((String) (jniCode.elementAt(2)));
		    native_post = ((String) (jniCode.elementAt(3)));
		    if (debug)
			System.err.println("---->" + native_pre + "/" + 
					   native_call + "/" + native_post);
		}
	    }
	    if (isConstructor) {
		ConstructorDefinition cdf = new ConstructorDefinition(metName, a);
		if (native_pre != null)
		    cdf.setNativePreGlue(native_pre);
		if (native_call != null)
		    cdf.setNativeGlue(native_call);
		if (native_post != null)
		    cdf.setNativePostGlue(native_post);
		addConstructorDefinition(cdf);
	    }
	    else {
		MethodDefinition cdf = new MethodDefinition(metName, retType, a);
		if (native_pre != null)
		    cdf.setNativePreGlue(native_pre);
		if (native_call != null)
		    cdf.setNativeGlue(native_call);
		if (native_post != null)
		    cdf.setNativePostGlue(native_post);
		addMethodDefinition(cdf);
	    }
	}
	catch (Exception e) {
	    System.err.println("parse define-func error: " + e.toString() + " " + v);
	    e.printStackTrace();
	}
    }
    
    public void parseDefineObject(Vector v) {
	TypeDefinition objectName;
	TypeDefinition inheritName;
	try {
	    objectName = getTypeDefinition((String) (v.elementAt(1)));
	    if (!objectName.getRawValue().equals("GtkObject")) {
		inheritName = getTypeDefinition((String)(((Vector)(v.elementAt(2))).firstElement()));
		addTypeDefinition(objectName);
		addTypeDefinition(inheritName);
		ClassDefinition cdf = new ClassDefinition(currentModule, modules, 
							  objectName, inheritName);
		if (v.size() == 4) {
		    Vector fields = (Vector) (v.elementAt(3));
		    Enumeration efields = fields.elements();
		    efields.nextElement(); // skipping "field"
		    
		    for (; efields.hasMoreElements();) {
			System.err.println("Field is" + efields);
			Vector field = (Vector) (efields.nextElement());
			for (Enumeration efld = field.elements(); 
			     efld.hasMoreElements();) {
			    System.err.println("Field is" + field);
			    String typename = (String) efld.nextElement();
			    String fldname = (String) efld.nextElement();
			    TypeDefinition fldtype = getTypeDefinition(typename);
			    //(String)e3.nextElement());
			    System.err.println("FIELDS:" + typename + fldname);
			    cdf.addField(new FieldDefinition(fldname, fldtype, cdf));
			}
		    }
		}
		addClassDefinition(cdf);
	    }
	}
	catch (Exception e) {
	    System.err.println("parse define-object error: " + e.toString() + 
			       " " + v);
	    e.printStackTrace();
	}
    }
    
    public void parseFile(String filename) {
	StreamTokenizer st;
	if (filename != null) {
	    System.out.println(">> loading defs file: " + filename);
	    try {
		BufferedReader r = 
		    new BufferedReader(new InputStreamReader(new FileInputStream(filename)));
		st = new StreamTokenizer(r);
		st.commentChar((int) ';');
		st.wordChars((int) '_', (int) '_');
		st.wordChars((int) '=', (int) '=');
		st.wordChars((int) '0', (int) '9');
		st.wordChars((int) '*', (int) '*');
		st.eolIsSignificant(false);
		st.nextToken();
		try {
		    File enum_file_gdk = new File("other" + File.separator + 
						  "Gdk.enums");
		    if (enum_file_gdk.exists())
			getAllEnums("other" + File.separator + "Gdk.enums");
		    File enum_file_gtk = new File("other" + File.separator + 
						  "Gtk.enums");
		    if (enum_file_gtk.exists())
			getAllEnums("other" + File.separator + "Gtk.enums");
		    //      File enum_file_gnome =
		    //        new File("other"+File.separator+"Gnome.enums");
		    //      if (enum_file_gnome.exists())
		    //        getAllEnums("other"+File.separator+"Gnome.enums");
		    //if (currentModule.equals("Gnome")) {
		    //	enum_file_gdk.delete();
		    //	enum_file_gtk.delete();
		    //          enum_file_gnome.delete();
		    //}
		    PrintWriter enum_pw = 
			new PrintWriter(new BufferedWriter(new FileWriter("other" + File.separator + currentModule + ".enums")));
		    while (st.ttype != st.TT_EOF) {
			if (st.ttype == (int) '(') {
			    Vector v;
			    v = parseSExpr(st);
			    if ((v.firstElement().equals("define-enum")) || 
				(v.firstElement().equals("define-flags")))
				enum_pw.println(v.elementAt(1));
			    if (debug)
				System.err.println("Raw sexpr:" + v);
			    if (v.firstElement().equals("define-object"))
				parseDefineObject(v);
			    if (v.firstElement().equals("define-boxed"))
				parseDefineBoxed(v);
			    if (v.firstElement().equals("define-func"))
				parseDefineFunc(v);
			    if (v.firstElement().equals("define-enum"))
				parseDefineEnum(v);
			    if (v.firstElement().equals("define-string-enum"))
				parseDefineEnum(v);
			    if (v.firstElement().equals("define-flags"))
				parseDefineFlags(v);
			    if (v.firstElement().equals("import"))
				parseImport(v);
			}
			st.nextToken();
		    }
		    enum_pw.close();
		}
		catch (Exception e) {
		    System.err.println(e);
		}
	    }
	    catch (Exception e) {
		System.err.println("parsefile error: " + e);
		e.printStackTrace();
	    }
	}
	rebuildClasses();
    }
    
    public void parseImport(Vector v) {
	String packageName = "";
	try {
	    Enumeration e = v.elements();
	    e.nextElement(); // "import"
	    
	    String s = (String) e.nextElement();
	    packageName = s.substring(0, s.length() - 5); // removing .defs
	    // System.err.println(packageName);
	    modules.addElement(packageName);
	}
	catch (Exception e) {
	    System.err.println("parse import error: " + e.toString() + " " + v);
	}
    }
    
    /** Obsolete. We now recreate the complete sexpr tree */
    public void parseSExpFlat(Vector v, StreamTokenizer st) {
	try {
	    st.nextToken();
	    while (st.ttype != (int) ')') {
		if (st.ttype == (int) '(')
		    parseSExpFlat(v, st);
		else if ((st.ttype == st.TT_WORD))
		    v.addElement(st.sval);
		else
		    return;
		st.nextToken();
	    }
	}
	catch (IOException e) {
	    System.err.println("parse: " + e);
	}
    }
    
    public Vector parseSExpr(StreamTokenizer st) {
	Vector v = new Vector();
	try {
	    st.nextToken();
	    while (st.ttype != (int) ')') {
		if (st.ttype == (int) '(')
		    v.addElement(parseSExpr(st));
		else if ((st.ttype == st.TT_WORD))
		    v.addElement(st.sval);
		else if ((st.ttype == (int) '"'))
		    v.addElement(st.sval);
		else if ((st.ttype == (int) '*'))
		    //          v.addElement(st.sval);
		    System.out.println("\n\n*\n\n");
		else
		    System.err.println(">> unexpected data in sexpr " + 
				       st.sval + "/" + st.ttype + "/" + v);
		st.nextToken();
	    }
	}
	catch (IOException e) {
	    System.err.println("parseSExpr error: " + e);
	}
	return v;
    }
    
    public void rebuildClasses() {
	Vector matched = new Vector();
	Vector matchedc = new Vector();
	
	// generating correct method attribution
	
	System.out.println(">> building method lists");
	for (Enumeration e = classnames.elements(); e.hasMoreElements();) {
	    ClassDefinition cd = (ClassDefinition) e.nextElement();
	    TypeDefinition s = cd.getClassType();
	    for (Enumeration e1 = constantnames.elements(); e1.hasMoreElements();) {
		ConstructorDefinition m = (ConstructorDefinition) e1.nextElement();
		if (m.translate(currentModule, s)) {
		    cd.addConstructor(m);
		    matchedc.addElement(m);
		}
	    }
	    for (Enumeration e1 = methodnames.elements(); e1.hasMoreElements();) {
		MethodDefinition m = (MethodDefinition) e1.nextElement();
		if (m.translate(currentModule, s)) {
		    cd.addMethod(m);
		    matched.addElement(m);
		}
	    }
	    if (debug)
		System.err.println(">> Saving class and glue code...");
	    cd.saveFile();
	    for (Enumeration e1 = matchedc.elements(); e1.hasMoreElements();) {
		// System.err.println("*** Cleaning Constructor List");
		ConstructorDefinition m = (ConstructorDefinition) e1.nextElement();
		constantnames.removeElement(m);
	    }
	    for (Enumeration e1 = matched.elements(); e1.hasMoreElements();) {
		// System.err.println("*** Cleaning Method List");
		MethodDefinition m = (MethodDefinition) e1.nextElement();
		methodnames.removeElement(m);
	    }
	}
	for (Enumeration e1 = enumsnames.elements(); e1.hasMoreElements();) {
	    ((EnumDefinition) e1.nextElement()).saveFile();
	}
	System.err.println("Static class construction");
	ClassDefinition cdf = new ClassDefinition(currentModule, modules, 
						  currentModule, null);
	//      for (Enumeration e1 = enumsnames.elements(); e1.hasMoreElements();)
	//  {
	//      cdf.addEnum((EnumDefinition)e1.nextElement());
	//  }
	
	System.err.println("Static class construction: enum ok");
	for (Enumeration e1 = methodnames.elements(); e1.hasMoreElements();) {
	    MethodDefinition m = (MethodDefinition) e1.nextElement();
	    // System.err.println("*** MainClass Unmatched"+m);
	    m.translate(currentModule, null);
	    // System.err.println("*** MainClass translate ok");
	    cdf.addMethod(m);
	    // System.err.println("*** MainClass addmethod ok");
	    matched.addElement(m);
	}
	if (debug)
	    System.err.println(">> Saving static class...");
	cdf.saveFile();
    }
}
